/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.pdfbox.util;

import org.apache.pdfbox.cos.COSArray;
import org.apache.pdfbox.cos.COSBase;
import org.apache.pdfbox.cos.COSDictionary;
import org.apache.pdfbox.cos.COSName;
import org.apache.pdfbox.cos.COSNumber;

import org.apache.pdfbox.pdmodel.common.COSArrayList;
import org.apache.pdfbox.pdmodel.graphics.color.PDColorSpace;
import org.apache.pdfbox.pdmodel.graphics.color.PDColorSpaceFactory;

import java.io.IOException;
import java.util.List;
import java.util.Map;

/**
 * This contains all of the image parameters for in inlined image.
 *
 * @author <a href="mailto:ben@benlitchfield.com">Ben Litchfield</a>
 * @version $Revision: 1.4 $
 */
public class ImageParameters implements bse.util.StaticStringInterface
{
    private COSDictionary dictionary;

    /**
     * Constructor.
     */
    public ImageParameters()
    {
        dictionary = new COSDictionary();
    }

    /**
     * Constructor.
     *
     * @param params The image parameters.
     */
    public ImageParameters(final COSDictionary params )
    {
        dictionary = params;
    }

    /**
     * This will get the dictionary that stores the image parameters.
     *
     * @return The COS dictionary that stores the image parameters.
     */
    public COSDictionary getDictionary()
    {
        return dictionary;
    }

    private COSBase getCOSObject(final String abbreviatedName, final String name )
    {
        final COSBase retval = dictionary.getDictionaryObject( COSName.getPDFName( abbreviatedName ) );
        if( retval == null )
        {
            return dictionary.getDictionaryObject( COSName.getPDFName( name ) );
        }
        return retval;
    }

    private int getNumberOrNegativeOne( String abbreviatedName, String name )
    {
        final COSNumber number = (COSNumber)getCOSObject( abbreviatedName, name );
        if( number != null )
        {
            return number.intValue();
        }
        return -1;
    }

    /**
     * The bits per component of this image.  This will return -1 if one has not
     * been set.
     *
     * @return The number of bits per component.
     */
    public int getBitsPerComponent()
    {
        return getNumberOrNegativeOne( BPC, "BitsPerComponent" );
    }

    /**
     * Set the number of bits per component.
     *
     * @param bpc The number of bits per component.
     */
    public void setBitsPerComponent(final int bpc )
    {
        dictionary.setInt( COSName.getPDFName( BPC ), bpc );
    }


    /**
     * This will get the color space or null if none exists.
     *
     * @return The color space for this image.
     *
     * @throws IOException If there is an error getting the colorspace.
     */
    public PDColorSpace getColorSpace() throws IOException
    {
        return getColorSpace( null );
    }
    
    /**
     * This will get the color space or null if none exists.
     *
     * @param colorSpaces The ColorSpace dictionary from the current resources, if any.
     *
     * @return The color space for this image.
     *
     * @throws IOException If there is an error getting the colorspace.
     */
    public PDColorSpace getColorSpace(final Map colorSpaces ) throws IOException
    {
        final COSBase cs = getCOSObject( CS, "ColorSpace" );
        if( cs != null )
        {
            return PDColorSpaceFactory.createColorSpace( cs, colorSpaces );
        }
        return null;
    }

    /**
     * This will set the color space for this image.
     *
     * @param cs The color space for this image.
     */
    public void setColorSpace(final PDColorSpace cs )
    {
        COSBase base = null;
        if( cs != null )
        {
            base = cs.getCOSObject();
        }
        dictionary.setItem( COSName.getPDFName( CS ), base );
    }

    /**
     * The height of this image.  This will return -1 if one has not
     * been set.
     *
     * @return The height.
     */
    public int getHeight()
    {
        return getNumberOrNegativeOne( H, "Height" );
    }

    /**
     * Set the height of the image.
     *
     * @param h The height of the image.
     */
    public void setHeight(final int h )
    {
        dictionary.setInt( COSName.getPDFName( H ), h );
    }

    /**
     * The width of this image.  This will return -1 if one has not
     * been set.
     *
     * @return The width.
     */
    public int getWidth()
    {
        return getNumberOrNegativeOne( W, "Width" );
    }

    /**
     * Set the width of the image.
     *
     * @param w The width of the image.
     */
    public void setWidth(final int w )
    {
        dictionary.setInt( COSName.getPDFName( W ), w );
    }

    /**
     * This will get the list of filters that are associated with this stream.  Or
     * null if there are none.
     * @return A list of all encoding filters to apply to this stream.
     */
    public List getFilters()
    {
        final COSBase filters = dictionary.getDictionaryObject( new String[] {"Filter", F} );
        if( filters instanceof COSName )
        {
            final COSName name = (COSName)filters;
            return new COSArrayList( name.getName(), name, dictionary, COSName.FILTER );
        }
        else if( filters instanceof COSArray )
        {
            return COSArrayList.convertCOSNameCOSArrayToList( (COSArray)filters );
        }
        return null;
    }

    /**
     * This will set the filters that are part of this stream.
     *
     * @param filters The filters that are part of this stream.
     */
    public void setFilters(final List filters )
    {
        dictionary.setItem( "Filter", COSArrayList.convertStringListToCOSNameCOSArray( filters ) );
    }
}